import random as rd
import numpy as np
from math import sqrt

### Partie I (juste pour qu'on puisse tester les fonctions des parties suivantes)

## A.1
def genePI(n,c):
    zone=[]
    for i in range(c+1):
        for j in range(c+1): zone.append((i,j))
    # en sortie, zone est la liste de tous les points possibles
    L=rd.sample(zone,n) # choix aléatoire de n éléments distincts dans zone
    return np.array(L) # conversion en tableau
    
## A.2
def distance(P,Q):
    x,y = P
    z,t = Q
    return sqrt((x-z)**2+(y-t)**2)
    
def calculdis(PI):
    n=len(PI)
    D=np.zeros((n+1,n+1)) # matrice remplie de 0
    R=(0,0) # position initiale du robot, pour fixer les idées
    for i in range(n):
        for j in range(i): # car la matrice sera symétrique
            D[i,j]=D[j,i]=distance(PI[i],PI[j])
        D[i,n]=D[n,i]=distance(PI[i],R)
        # remplissage de la dernière ligne et colonne
    return D

### Partie II 
   
## A.1
def long(chem,D):
    l=0
    i=len(D)-1  # point de départ: position du robot
    for j in chem:
        l+=D[i,j]
        i=j  # nouveau point de départ
    return round(l,2) # arrondi à 2 décimales
    
## A.2
def normaliser(che,n):
    test=[True]*n 
    # liste des disponibilités des points du chemin
    Che=[]  # sera le futur chemin normalisé
    for i in che:
        if i<n and test[i]: # si i est un point du chemin et est disponible
            Che.append(i)
            test[i]=False  # ce point n'est plus disponible
    # en sortie, che est dédoublonné ; il reste à le compléter avec les points disponibles restants
    for i in range(n):
        if test[i]: Che.append(i)
    return Che
    
## B.1
# Chemin = permutation des n points, donc il y a n! chemins.

## B.2
# Pour n=20: 20! est de l'ordre de 10**18 ; à raison de 10**(-6) seconde pour tester un chemin, il faudra 10**12 secondes, soit 30 000 ans.

## C.1
def ppv(p,D,test):
    # p: point, D:table des distances, test: liste de booléens correspondants aux points pas encore visités
    n=len(D)-1  # nombre de points
    mini=-1 
    for k in range(n):
        if test[k] and (mini<0 or D[p,k]<D[p,mini]): mini=k
        # on s'assure que le point k n'est pas encore vu et plus proche que le point mini provisoire; le test mini<0 doit précéder D[p,mini] sinon on risque une erreur à l'exécution
    return mini
    
def PPV(D):
    n=len(D)-1 
    chem=[]
    test=[True]*n+[False] # le point n est la position du robot, on y est
    p=n
    for i in range(n):
        p=ppv(p,D,test) # plus proche voisin de p
        chem.append(p)
        test[p]=False # p n'est plus disponible
    return chem
    
## C.2
# ppv a pour complexité O(n) car 1 boucle de longueur n. PPV a des initialisations en O(n) et 1 boucle de longueur n avec ppv à chaque tour, donc O(n²). D est en O(n²) aussi, donc au total O(n²).

## C.3
# Si le robot est en (0,2000), le chemin donné par PPV est [1,0,2] de longueur 11 mètres, alors que le chemin minimal est [0,1,2] de longueur 9 mètres.

### Partie III

## A
def creepop(m,D):
    P=[] # future population
    n=len(D)-1
    points=list(range(n)) # chemin initial: entiers de 0 à n-1
    for i in range(m):
        rd.shuffle(points) # chemin obtenu par mélange en place du précédent
        chem=list(points) # copie inerte de ce chemin
        l=long(chem,D) # sa longueur
        P.append((l,chem))
    return P
    # P.append((l,points)) ne fonctionne pas car les listes sont passées par adresse: points sera alors le dernier chemin créé et non celui qu'on vient de construire (à vérifier !)
    
## B
def reduire(P):
    m=len(P)
    P.sort()  # tri en place par longueur, car l'ordre sur les couples est lexicographique
    for i in range(m//2): P.pop() # on retire les m//2 derniers
    
## C.1
def muter(che):
    n=len(che)
    i,j = rd.sample(range(n),2) # 2 numéros aléatoires distincts
    che[i],che[j] = che[j],che[i]
    # procédure : ne retourne rien.
    
## C.2
def Muter(P,p,D):
    m=len(P)
    for i in range(m):
        if rd.random()<p: # ce qui arrive avec la probabilité p
            che=P[i][1] # P[i] est un couple
            muter(che) # mutation en place (procédure)
            l=long(che,D)
            P[i]=(l,che)
    # procédure : ne retourne rien.
            
## D.1
def croiser(C1,C2):
    n=len(C1)
    p=n//2
    return normaliser(C1[:p]+C2[p:],n)
    # on obtient bien un chemin valide et de longueur n.
    
## D.2
def newG(P,D):
    m=len(P)
    for i in range(-1,m-1):
        # au départ P[i] est le dernier couple, P[i+1] est le premier
        C1=P[i][1] # chemin correspondant au couple P[i]
        C2=P[i+1][1]
        chem=croiser(C1,C2)
        l=long(chem,D)
        P.append((l,chem))
    # en sortie P est de longueur 2m.
        
## E.1
def algoGen(PI,m,p,g):
    D=calculdis(PI)
    P=creepop(m,D)
    for i in range(g):
        reduire(P) # P est alors de longueur m/2
        newG(P,D) # P est alors de longueur m
        Muter(P,p,D)
    imin=0
    for i in range(1,m):
        if P[i][0]<P[imin][0]: imin=i # comparaison des longueurs
    return P[imin] # le meilleur de la population finale.
    
## E.2
# Oui: si la génération n a trouvé le meilleur chemin et que c'est celui-ci qui est muté ensuite, le résultat de la génération n+1 sera moins bon.
# Pour éviter ceci, on pourrait comparer la longueur du mutant à la longueur avant mutation, et ne garder le mutant que s'il est meilleur.

## E.3
# On peut envisager de calculer le min à chaque génération, et de stopper lorsque ce min ne bouge plus au cours de p générations successives. Inconvénients : on augmente le temps de calcul à cause des min; si p est trop petit, on peut n'avoir obtenu qu'un min local.
# Autre option: faire 5 générations, calculer le min et le comparer à celui donné par PPV ; on continue par paquets de 5 jusqu'à obtenir mieux que PPV ; ensuite faire encore 10 générations et stopper.Inconvénients: ne marchera pas si PPV est vraiment mauvais.
# Un hybride des 2 précédents...
